"""Error tracking service API and setup."""


import re

import sentry_sdk
from pyramid.settings import asbool
from sentry_sdk.integrations.logging import ignore_logger
from sentry_sdk.integrations.pyramid import PyramidIntegration

from h_pyramid_sentry.event_filter import get_before_send


def report_exception(exc=None):
    """Report an exception to the error tracking service.

    If the given ``exc`` is :obj:`None` then the most recently raised exception
    will be reported.

    :arg exc: the exception to report
    :type exc: :class:`Exception`, :obj:`None`, or a :func:`sys.exc_info` tuple
    """
    sentry_sdk.capture_exception(exc)


DEFAULT_OPTIONS = {"send_default_pii": True, "integrations": [PyramidIntegration()]}
OPTION_PATTERN = re.compile(r"^h_pyramid_sentry\.init\.(.*)$")


def includeme(config):
    """Set up the error tracking service."""
    filters = [
        # Allow functions to be passed as strings: e.g. "module.function"
        config.maybe_dotted(_filter)
        for _filter in config.registry.settings.get("h_pyramid_sentry.filters", [])
    ]

    if asbool(config.registry.settings.get("h_pyramid_sentry.retry_support")):
        # pylint:disable=import-outside-toplevel
        # This is here to lazy load only when required
        from h_pyramid_sentry.filters.pyramid import is_retryable_error

        filters.append(is_retryable_error)
        config.scan("h_pyramid_sentry.subscribers")

    init_options = {**DEFAULT_OPTIONS}
    for key, value in config.registry.settings.items():
        match = OPTION_PATTERN.match(key)
        if match:
            init_options[match.group(1)] = value

    if asbool(config.registry.settings.get("h_pyramid_sentry.celery_support")):
        # pylint:disable=import-outside-toplevel
        # This is here to lazy load only when required
        from sentry_sdk.integrations.celery import CeleryIntegration

        init_options["integrations"].append(CeleryIntegration())

    # exc_logger (which comes from
    # https://docs.pylonsproject.org/projects/pyramid_exclog/) logs all
    # exceptions with log level ERROR, and then all of these get picked up by
    # sentry_sdk's enabled-by-default logging integration
    # (https://docs.sentry.io/platforms/python/logging/).
    #
    # This means that:
    #
    # 1. All exceptions get reported via exc_logger instead of via sentry_sdk's
    # normal route (of intercepting raised exceptions)
    #
    # 2. Some handled exceptions that wouldn't have been reported via the
    # normal route now do get reported because exc_logger logs them as errors.
    #
    # We don't want either of these two things to happen so tell sentry_sdk's
    # logging integration to ignore exc_logger (see
    # https://docs.sentry.io/platforms/python/logging/#ignoring-a-logger).
    #
    # If pyramd_exclog isn't installed this will just have no effect.
    ignore_logger("exc_logger")

    sentry_sdk.init(before_send=get_before_send(filters), **init_options)



"""An object which filters events."""
import logging

from h_pyramid_sentry.event import Event
from h_pyramid_sentry.exceptions import FilterNotCallableError

LOG = logging.getLogger(__name__)
LOG_MESSAGE_PREFIX = "Filtering out Sentry event"
LOG_MESSAGE_TEMPLATE = f"{LOG_MESSAGE_PREFIX}: %s"


def get_before_send(filters=None):
    """Return a function which applies the given filters to events.

    Each time an event (for example an uncaught exception or a logged
    error) that would be reported to Sentry happens, ``sentry_sdk`` calls
    this function passing the event first.

    If this function returns ``event_dict`` then the event will be reported
    to Sentry. If this function returns ``None`` the event won't be
    reported.

    See https://docs.sentry.io/error-reporting/configuration/filtering/
    """

    filters = filters or []

    for filter_function in filters:
        if not callable(filter_function):
            raise FilterNotCallableError(filter_function)

    def _before_send(event_dict, hint_dict):
        event = Event(event_dict, hint_dict)

        if any(filter_function(event) for filter_function in filters):
            LOG.info(LOG_MESSAGE_TEMPLATE, hint_dict)
            return None

        return event_dict

    return _before_send



"""Decorator class for Sentry events.

Centralizes Sentry event parsing in one reusable place. Sentry events don't
have a very convenient interface in the form that sentry_sdk sends them to
us in.

They're dicts of dicts and tuples, and the structure changes depending on
whether the event is an exception raised or an error logged. This class
provides a more convenient interface to the interesting properties of events.
"""


class Event:
    """A decorator for a sentry_sdk event."""

    def __init__(self, event, hint):
        self._event = event
        self._hint = hint

    @property
    def event(self):
        """Get the raw event dict as passed to us by sentry_sdk."""
        return self._event

    @property
    def hint(self):
        """Get the raw hint dict as passed to us by sentry_sdk."""
        return self._hint

    @property
    def logger(self):
        """Get the name of the logger that logged the error message.

        A string, or None of this isn't a logger event.
        """
        return self.event.get("logger")

    @property
    def message(self):
        """Get the message that was logged.

        A string, or None of this isn't a logger event.
        """
        return self.event.get("logentry", {}).get("message")

    @property
    def exception(self):
        """Get the actual exception object that was raised.

        None if this isn't an exception event.
        """
        return self.hint.get("exc_info", (None, None, None))[1]



"""All exceptions defined for h_pyramid_sentry."""


class FilterNotCallableError(ValueError):
    """Indicates we were passed a non-callable object as a filter."""

    def __init__(self, _filter):
        self.filter = _filter

        super().__init__(f"Filter function is not callable: {type(_filter)}")



"""A pyramid subscriber to add extra info to retryable events."""
import traceback

import sentry_sdk
from pyramid.events import subscriber
from pyramid_retry import IBeforeRetry


@subscriber(IBeforeRetry)
def add_retryable_error_to_sentry_context(event):
    """Add information about a retryable error to the Sentry context.

    If a request raises a retryable error then, if we're not already on this
    request's last retry attempt, pyramid_retry calls this IBeforeRetry
    subscriber function before it retries the request.

    This function adds some information to the Sentry context about the
    retryable error that was raised by the failed attempt at the request.

    If a future retry of this request raises a non-retryable error, or if we
    use up the maximum number of retry attempts, so that we ultimately end up
    sending an error response back to the client, then the information that
    this function adds to the Sentry context will be included in the event that
    is ultimately reported to Sentry.

    This makes debugging errors easier because you can see information about
    previous failed attempts at the request, as well as the usual information
    about the last exception that was raised by the final attempt.

    If a future retry of this request succeeds and we ultimately end up sending
    a successful response back to the client then nothing will be reported to
    Sentry.

    See:

    * https://docs.pylonsproject.org/projects/pyramid-retry/en/latest/#receiving-retry-notifications
    * https://docs.sentry.io/platforms/python/#extra-context
    """
    attempt = event.environ.get("retry.attempt")
    attempts = event.environ.get("retry.attempts")
    exception = event.exception

    if attempt is None or attempts is None:
        return

    # The first attempt is numbered 0. Let's make it more human.
    attempt = attempt + 1

    traceback_str = "".join(
        traceback.format_exception(type(exception), exception, exception.__traceback__)
    )

    # Sentry event variables are limited to 512 characters long, otherwise they
    # get truncated: https://docs.sentry.io/development/sdk-dev/data-handling/
    # We don't want the end of our stack trace to get truncated because the end
    # of a stack trace is the most useful part of it, so truncate the start
    # ourselves instead.
    if len(traceback_str) > 512:
        traceback_str = "..." + traceback_str[-509:]

    with sentry_sdk.configure_scope() as scope:
        scope.set_extra(
            f"Exception from attempt {attempt}/{attempts}",
            repr(exception) or str(exception),
        )
        scope.set_extra(
            f"End of traceback from attempt {attempt}/{attempts}", traceback_str
        )



"""Filters specifically for Pyramid."""

# Import is_error_retryable indirectly to make importing from us less
# confusing as we have a very similarly named function
import pyramid_retry
from pyramid.threadlocal import get_current_request


def is_retryable_error(event):
    """Filter exceptions from requests that are going to be retried.

    If a request raises a retryable error, so pyramid_retry automatically
    retries that request, and a subsequent retry succeeds and we end up sending
    a successful response back to the client, then we don't want to report
    anything to Sentry.

    See:

    https://docs.pylonsproject.org/projects/pyramid-retry/en/latest/api.html#pyramid_retry.is_error_retryable
    """
    request = get_current_request()

    if request is None:
        # get_current_request() returns None if we're outside of Pyramid's
        # request context. This happens when sentry-sdk's Pyramid integration
        # catches an exception that Pyramid is going to raise to the WSGI
        # server (Gunicorn, in our case).
        # Always allow these uncaught exceptions to be reported to Sentry:
        return False

    if pyramid_retry.is_error_retryable(request, event.exception):
        # Don't report the exception to Sentry if the request is going to be
        # re-tried.
        #
        # Note: is_error_retryable() returns *False* if the error is marked as
        # retryable but the request is on its last attempt and is not going to
        # be re-tried again, so if a request fails with a retryable exception
        # on its last attempt (and we send back an error response to the
        # client) that exception will still be reported.
        return True

    return False



import os

from setuptools import find_packages, setup
from setuptools.config import read_configuration


class Package:
    def __init__(self, config):
        metadata = config['metadata']
        options = config['options']

        self.options = options
        self.name = metadata['name']
        self.version = metadata['version']

    def tests_require(self):
        return self.options['tests_require'] + self.options['install_requires']

    def read_egg_version(self):
        pkg_info_file = None
        # PKG-INFO can be in different places depending on whether we are a
        # source distribution or a checked out copy etc.
        for location in [
                'PKG-INFO',
                'src/' + self.name + ".egg-info/PKG-INFO",
                self.name + ".egg-info/PKG-INFO",
            ]:
            if os.path.isfile(location):
                pkg_info_file = location

        if not pkg_info_file:
            return None

        with open(pkg_info_file) as fh:
            for line in fh:
                if line.startswith("Version"):
                    return line.strip().split("Version: ")[-1]

    def get_version(self, build_var="BUILD"):
        """Gets a version reading from the specified environment variable

        This expects the variable to contain something like:

         * refs/heads/<branch_name> - This will be ignored
         * refs/tags/v<version> - This will be used if it matches our major
                                  minor number
         * v<version> - This will be used if it matches our major minor number

        If this is not present then we will read from the .egg-info/ data if
        possible.

        Finally a fallback development version is provided.

        :param build_var: The enviroment to
        :return: A version string
        """
        # If we have a build argument we should honour it if we can
        build = os.environ.get(build_var)
        if build:
            if build.startswith('refs/heads/'):
                # We are being built via CI from a branch: we'll return a
                # dummy value marking this as an 'alpha' release
                return self.version + '.a0'

            if build.startswith('refs/tags/'):
                # We are being built via CI from a tag: strip the refs stuff
                build = build.replace('refs/tags/', '')

            start = 'v' + self.version + '.'
            if not build.startswith(start):
                raise ValueError(f'Expected build to be "{start}*", got "{build}"')

            return self.version + "." + build[len(start):]

        # If not, we should try and read it from the .egg-info/ data

        # We need to do this for source distributions, as setup.py is re-run
        # when installed this way, and we would always get 'dev0' as the
        # version wheels and binary installs don't work this way and read
        # from PKG-INFO for them selves
        egg_version = self.read_egg_version()
        if egg_version:
            return egg_version

        # Otherwise create a 'dev' build which will be counted by pip as
        # 'later' than the major version no matter what
        return self.version + ".dev0"


package = Package(read_configuration('setup.cfg'))

setup(
    # Metadata
    # https://docs.python.org/3/distutils/setupscript.html#additional-meta-data
    version=package.get_version(),

    # Contents and dependencies
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    # Read the MANIFEST.in
    include_package_data=True,

    # Add support for pip install .[tests]
    extras_require={"tests": package.tests_require()},

    tests_require=package.tests_require(),
)
